# Copyright (c) 2023 Graphcore Ltd. All rights reserved.
import chex
import numpy as np
import numpy.testing as npt
from absl.testing import parameterized
from jax import lax

from jax_scalify.core import Array, ScaledArray, debug_callback, find_registered_scaled_op, scaled_array, scalify
from jax_scalify.lax import (
    scaled_broadcast_in_dim,
    scaled_concatenate,
    scaled_convert_element_type,
    scaled_is_finite,
    scaled_pad,
    scaled_reduce_precision,
    scaled_reshape,
    scaled_rev,
    scaled_select_n,
    scaled_sign,
    scaled_slice,
    scaled_transpose,
)


class ScaledTranslationPrimitivesTests(chex.TestCase):
    def setUp(self):
        super().setUp()
        # Use random state for reproducibility!
        self.rs = np.random.RandomState(42)

    @chex.variants(with_jit=True, without_jit=True)
    def test__scaled_debug_callback__proper_forwarding(self):
        host_values = []

        def callback(*args):
            for v in args:
                host_values.append(v)

        def fn(a):
            # NOTE: multiplying by a power of 2 to simplify test.
            debug_callback(callback, a, a * 4)
            return a

        x = scaled_array(self.rs.rand(5), 2, dtype=np.float16)
        fn = self.variant(scalify(fn))
        fn(x)

        assert len(host_values) == 2
        for sv in host_values:
            assert isinstance(sv, ScaledArray)
            npt.assert_array_equal(sv.data, x.data)
        npt.assert_array_equal(host_values[0].scale, x.scale)
        npt.assert_array_equal(host_values[1].scale, x.scale * 4)

    def test__scaled_broadcast_in_dim__proper_scaling(self):
        x = scaled_array(self.rs.rand(5), 2, dtype=np.float32)
        z = scaled_broadcast_in_dim(x, shape=(5, 1), broadcast_dimensions=(0,))
        assert isinstance(z, ScaledArray)
        npt.assert_array_equal(z.scale, x.scale)
        npt.assert_array_almost_equal(z.data, x.data.reshape((5, 1)))

    def test__scaled_reshape__proper_scaling(self):
        x = scaled_array(self.rs.rand(8), 2, dtype=np.float32)
        z = scaled_reshape(x, new_sizes=(4, 2), dimensions=None)
        assert isinstance(z, ScaledArray)
        npt.assert_array_equal(z.scale, x.scale)
        npt.assert_array_almost_equal(z.data, x.data.reshape((4, 2)))

    def test__scaled_concatenate__proper_scaling(self):
        x = scaled_array(self.rs.rand(2, 3), 0.5, dtype=np.float16)
        y = scaled_array(self.rs.rand(5, 3), 2, dtype=np.float16)
        z = scaled_concatenate([x, y], dimension=0)
        assert isinstance(z, ScaledArray)
        assert z.dtype == x.dtype
        npt.assert_array_equal(z.scale, y.scale)
        npt.assert_array_almost_equal(z, lax.concatenate([np.asarray(x), np.asarray(y)], dimension=0))

    def test__scaled_concatenate__zero_input_scales(self):
        x = scaled_array(self.rs.rand(2, 3), 0.0, dtype=np.float16)
        y = scaled_array(self.rs.rand(5, 3), 0.0, dtype=np.float16)
        z = scaled_concatenate([x, y], dimension=0)
        assert isinstance(z, ScaledArray)
        assert z.dtype == x.dtype
        npt.assert_array_equal(z.scale, 0)
        npt.assert_array_almost_equal(z, lax.concatenate([np.asarray(x), np.asarray(y)], dimension=0))

    def test__scaled_convert_element_type__proper_scaling(self):
        x = scaled_array(self.rs.rand(5), 2, dtype=np.float32)
        z = scaled_convert_element_type(x, new_dtype=np.float16)
        assert isinstance(z, ScaledArray)
        npt.assert_array_equal(z.scale, x.scale)
        npt.assert_array_almost_equal(z.data, x.data.astype(z.dtype))

    def test__scaled_transpose__proper_scaling(self):
        x = scaled_array(self.rs.rand(3, 5), 2, dtype=np.float32)
        z = scaled_transpose(x, (1, 0))
        assert isinstance(z, ScaledArray)
        assert z.scale == x.scale
        npt.assert_array_almost_equal(z.data, x.data.T)

    def test__scaled_sign__proper_scaling(self):
        x = scaled_array(self.rs.rand(3, 5), 2, dtype=np.float32)
        z = scaled_sign(x)
        assert isinstance(z, ScaledArray)
        assert z.scale == 1
        npt.assert_array_equal(z.data, lax.sign(x.data))

    def test__scaled_rev__proper_scaling(self):
        x = scaled_array(self.rs.rand(5), 2, dtype=np.float32)
        z = scaled_rev(x, dimensions=(0,))
        assert isinstance(z, ScaledArray)
        assert z.scale == x.scale
        npt.assert_array_almost_equal(z.data, x.data[::-1])

    def test__scaled_pad__proper_scaling(self):
        x = scaled_array(self.rs.rand(5), 2, dtype=np.float32)
        z = scaled_pad(x, 0.0, [(1, 2, 0)])
        assert isinstance(z, ScaledArray)
        assert z.scale == x.scale
        npt.assert_array_almost_equal(z.data, lax.pad(x.data, 0.0, [(1, 2, 0)]))

    def test__scaled_reduce_precision__proper_result(self):
        x = scaled_array(self.rs.rand(3, 5), 2, dtype=np.float16)
        # Reduction to pseudo FP8 format.
        z = scaled_reduce_precision(x, exponent_bits=4, mantissa_bits=3)
        assert isinstance(z, ScaledArray)
        assert z.dtype == x.dtype
        assert z.scale == x.scale
        npt.assert_array_almost_equal(z.data, lax.reduce_precision(x.data, exponent_bits=4, mantissa_bits=3))

    def test__scaled_slice__proper_scaling(self):
        x = scaled_array(self.rs.rand(5), 2, dtype=np.float32)
        z = scaled_slice(x, (1,), (4,), (2,))
        assert isinstance(z, ScaledArray)
        assert z.scale == x.scale
        npt.assert_array_almost_equal(z.data, x.data[1:4:2])

    @parameterized.parameters({"prim": lax.argmax_p}, {"prim": lax.argmin_p})
    def test__scaled_argminmax__proper_scaling(self, prim):
        x = scaled_array(self.rs.rand(5), 2, dtype=np.float32)
        expected_out = prim.bind(x.to_array(), axes=(0,), index_dtype=np.int32)
        scaled_translation, _ = find_registered_scaled_op(prim)
        out = scaled_translation(x, axes=(0,), index_dtype=np.int32)
        assert isinstance(out, Array)
        npt.assert_array_equal(out, expected_out)


class ScaledTranslationBooleanPrimitivesTests(chex.TestCase):
    def setUp(self):
        super().setUp()
        # Use random state for reproducibility!
        self.rs = np.random.RandomState(42)

    @parameterized.parameters(
        {"val": scaled_array([2, 3], 2.0, dtype=np.float32), "expected_out": [True, True]},
        # Supporting `int` scale as well.
        {"val": scaled_array([2, np.inf], 2, dtype=np.float32), "expected_out": [True, False]},
        {"val": scaled_array([2, 3], np.nan, dtype=np.float32), "expected_out": [False, False]},
        {"val": scaled_array([np.nan, 3], 3.0, dtype=np.float32), "expected_out": [False, True]},
    )
    def test__scaled_is_finite__proper_result(self, val, expected_out):
        out = scaled_is_finite(val)
        assert isinstance(out, Array)
        assert out.dtype == np.bool_
        npt.assert_array_equal(out, expected_out)

    @parameterized.parameters(
        {"bool_prim": lax.eq_p},
        {"bool_prim": lax.ne_p},
        {"bool_prim": lax.lt_p},
        {"bool_prim": lax.le_p},
        {"bool_prim": lax.gt_p},
        {"bool_prim": lax.ge_p},
    )
    def test__scaled_boolean_binary_op__proper_result(self, bool_prim):
        lhs = scaled_array(self.rs.rand(5), 2.0, dtype=np.float32)
        rhs = scaled_array(self.rs.rand(5), 3.0, dtype=np.float32)
        scaled_bool_op, _ = find_registered_scaled_op(bool_prim)
        out0 = scaled_bool_op(lhs, rhs)
        out1 = scaled_bool_op(lhs, lhs)
        assert isinstance(out0, Array)
        assert out0.dtype == np.bool_
        npt.assert_array_equal(out0, bool_prim.bind(lhs.to_array(), rhs.to_array()))
        npt.assert_array_equal(out1, bool_prim.bind(lhs.to_array(), lhs.to_array()))

    def test__scaled_select_n__proper_result(self):
        mask = self.rs.rand(5) > 0.5
        lhs = scaled_array(self.rs.rand(5), 2.0, dtype=np.float32)
        rhs = scaled_array(self.rs.rand(5), 4.0, dtype=np.float32)
        out = scaled_select_n(mask, lhs, rhs)
        assert isinstance(out, ScaledArray)
        assert out.dtype == np.float32
        # Max scale used.
        npt.assert_almost_equal(out.scale, 4)
        npt.assert_array_equal(out, np.where(mask, rhs, lhs))

    @parameterized.parameters(
        {"scale": 0.25},
        {"scale": 8.0},
    )
    def test__scaled_select__relu_grad_example(self, scale):
        @scalify
        def relu_grad(g):
            return lax.select(g > 0, g, lax.full_like(g, 0))

        # Gradient with some scale.
        gin = scaled_array([1.0, 0.5], np.float32(scale), dtype=np.float32)
        gout = relu_grad(gin)
        # Same scale should be propagated to gradient output.
        assert isinstance(gout, ScaledArray)
        npt.assert_array_equal(gout.scale, gin.scale)



# Copyright (c) 2023 Graphcore Ltd. All rights reserved.
import chex
import numpy as np
import numpy.testing as npt
from absl.testing import parameterized

from jax_scalify.core import ScaledArray, scaled_array
from jax_scalify.ops import dynamic_rescale_l1, dynamic_rescale_l2, dynamic_rescale_max


class DynamicRescaleOpsTests(chex.TestCase):
    def test__dynamic_rescale_max__proper_max_rescale_pow2_rounding(self):
        arr_in = scaled_array([2, -3], np.float16(4), dtype=np.float16)
        arr_out = dynamic_rescale_max(arr_in)

        assert isinstance(arr_out, ScaledArray)
        assert arr_out.dtype == arr_in.dtype
        npt.assert_array_equal(arr_out.scale, np.float16(8))
        npt.assert_array_equal(arr_out, arr_in)

    def test__dynamic_rescale_l1__proper_l1_rescale_pow2_rounding(self):
        # L1 norm = 2
        arr_in = scaled_array([1, -6], np.float16(4), dtype=np.float16)
        arr_out = dynamic_rescale_l1(arr_in)

        assert isinstance(arr_out, ScaledArray)
        assert arr_out.dtype == arr_in.dtype
        npt.assert_array_equal(arr_out.scale, np.float16(8))
        npt.assert_array_equal(arr_out, arr_in)

    def test__dynamic_rescale_l2__proper_max_rescale_pow2_rounding(self):
        # L2 norm = 8.945
        arr_in = scaled_array([4, -8], np.float16(4), dtype=np.float16)
        arr_out = dynamic_rescale_l2(arr_in)

        assert isinstance(arr_out, ScaledArray)
        assert arr_out.dtype == arr_in.dtype
        npt.assert_array_equal(arr_out.scale, np.float16(16))
        npt.assert_array_equal(arr_out, arr_in)

    @parameterized.parameters(
        {"dynamic_rescale_fn": dynamic_rescale_max},
        {"dynamic_rescale_fn": dynamic_rescale_l1},
        {"dynamic_rescale_fn": dynamic_rescale_l2},
    )
    def test__dynamic_rescale__epsilon_norm_value(self, dynamic_rescale_fn):
        arr_in = scaled_array([0, 0], np.float32(1), dtype=np.float16)
        arr_out = dynamic_rescale_fn(arr_in)
        # Rough bounds on the epsilon value.
        assert arr_out.scale > 0.0
        assert arr_out.scale < 0.001



# Copyright (c) 2023 Graphcore Ltd. All rights reserved.
import chex
import numpy as np
import numpy.testing as npt
from absl.testing import parameterized
from jax import lax

from jax_scalify.core import scaled_array, scalify


class ScaledScipyHighLevelMethodsTests(chex.TestCase):
    def setUp(self):
        super().setUp()
        # Use random state for reproducibility!
        self.rs = np.random.RandomState(42)

    def test__lax_full_like__zero_scale(self):
        def fn(a):
            return lax.full_like(a, 0)

        a = scaled_array(np.random.rand(3, 5).astype(np.float32), np.float32(1))
        scalify(fn)(a)
        # FIMXE/TODO: what should be the expected result?

    @chex.variants(with_jit=False, without_jit=True)
    @parameterized.parameters(
        {"dtype": np.float32},
        {"dtype": np.float16},
    )
    def test__scipy_logsumexp__accurate_scaled_op(self, dtype):
        from jax.scipy.special import logsumexp

        input_scaled = scaled_array(self.rs.rand(10), 4.0, dtype=dtype)
        # JAX `logsumexp` Jaxpr is a non-trivial graph!
        out_scaled = self.variant(scalify(logsumexp))(input_scaled)
        out_expected = logsumexp(np.asarray(input_scaled))
        assert out_scaled.dtype == out_expected.dtype
        # Proper accuracy + keep the same scale.
        npt.assert_array_equal(out_scaled.scale, input_scaled.scale)
        npt.assert_array_almost_equal(out_scaled, out_expected, decimal=5)



# Copyright (c) 2023 Graphcore Ltd. All rights reserved.
import chex
import jax
import jax.numpy as jnp
import numpy as np

from jax_scalify.core import ScaledArray, scaled_array, scalify


class ScaledJaxNumpyFunctions(chex.TestCase):
    def setUp(self):
        super().setUp()
        # Use random state for reproducibility!
        self.rs = np.random.RandomState(42)

    @chex.variants(with_jit=True, without_jit=False)
    def test__numpy_mean__proper_gradient_scale_propagation(self):
        def mean_fn(x):
            # Taking the square to "force" ScaledArray gradient.
            # Numpy mean constant rescaling creating trouble on backward pass!
            return jax.grad(lambda v: jnp.mean(v * v))(x)

        # size = 8 * 16
        input_scaled = scaled_array(self.rs.rand(8, 16).astype(np.float32), np.float32(1))
        output_grad_scaled = self.variant(scalify(mean_fn))(input_scaled)

        assert isinstance(output_grad_scaled, ScaledArray)
        # Proper scale propagation on the backward pass (rough interval)
        assert np.std(output_grad_scaled.data) >= 0.25
        assert np.std(output_grad_scaled.data) <= 1.0
        # "small" scale.
        assert output_grad_scaled.scale <= 0.01



# Copyright (c) 2023 Graphcore Ltd. All rights reserved.
from functools import partial
from typing import Sequence

import jax

from jax_scalify.core import Array, debug_callback


@partial(jax.custom_vjp, nondiff_argnums=(0,))
def debug_callback_grad(f, *args):
    """Custom callback, called on gradients."""
    return args


def debug_callback_grad_fwd(f, *args):
    return args, None


def debug_callback_grad_bwd(f, _, args_grad):
    debug_callback(f, *args_grad)
    return args_grad


debug_callback_grad.defvjp(debug_callback_grad_fwd, debug_callback_grad_bwd)


def debug_print(fmt: str, *args: Array) -> Sequence[Array]:
    """Debug print of a collection of tensors."""
    debug_callback(lambda *args: print(fmt.format(*args)), *args)
    return args


def debug_print_grad(fmt: str, *args: Array) -> Sequence[Array]:
    """Debug print of gradients of a collection of tensors."""
    return debug_callback_grad(lambda *args: print(fmt.format(*args)), *args)



# Copyright (c) 2023 Graphcore Ltd. All rights reserved.
from .datatype import (  # noqa: F401
    DTypeLike,
    ScaledArray,
    Shape,
    as_scaled_array,
    asarray,
    get_scale_dtype,
    is_scaled_leaf,
    is_static_anyscale,
    is_static_one_scalar,
    is_static_zero,
    make_scaled_scalar,
    scaled_array,
)
from .debug import debug_callback  # noqa: F401
from .interpreters import (  # noqa: F401
    ScaledPrimitiveType,
    ScalifyConfig,
    find_registered_scaled_op,
    get_scalify_config,
    register_scaled_lax_op,
    register_scaled_op,
    scalify,
)
from .pow2 import Pow2RoundMode, pow2_decompose, pow2_round, pow2_round_down, pow2_round_up  # noqa: F401
from .typing import Array, ArrayTypes, get_numpy_api  # noqa: F401
from .utils import safe_div, safe_reciprocal  # noqa: F401



# Copyright (c) 2023 Graphcore Ltd. All rights reserved.
import logging
from enum import IntEnum
from functools import partial
from typing import Any, Dict, Optional, Sequence, Tuple, Union

import numpy as np
from jax import core
from jax.interpreters import mlir
from jax.interpreters.mlir import LoweringRuleContext, ir
from numpy.typing import DTypeLike, NDArray

from .typing import Array, get_numpy_api

# Exponent bits masking.
_exponent_bits_mask: Dict[Any, NDArray[Any]] = {
    np.dtype(np.float16): np.packbits(np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0], dtype=np.uint8)).view(
        np.int16
    ),
    np.dtype(np.float32): np.packbits(
        np.array(
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1],
            dtype=np.uint8,
        )
    ).view(np.int32),
    np.dtype(np.float64): np.array(np.inf, np.float64).view(np.int64),
}
"""Exponents bit masking: explicit bitmask to keep only exponent bits in floating point values.

NOTE: normally should also correspond to `np.inf` value for FP16 and FP32.
"""


def pow2_decompose_round_down_impl(vin: Array, scale_dtype: DTypeLike) -> Array:
    """Pow-2 decompose with rounding down.

    Returns:
        (scale, vout) such that vin = scale * vout
    """
    np_api = get_numpy_api(vin)
    # Perform all computations in FP32, to support FP16 submormals.
    # NOTE: `jnp.frexp` is buggy for subnormals.
    dtype = np.dtype(np.float32)
    minval = np.finfo(dtype).smallest_normal
    exponent_mask = _exponent_bits_mask[dtype]
    intdtype = exponent_mask.dtype
    val = vin.astype(dtype)
    # Masking mantissa bits, keeping only the exponents ones.
    scale_pow2 = np_api.bitwise_and(val.view(intdtype), exponent_mask).view(val.dtype).reshape(val.shape)
    # Get the mantissa in float32. Make sure we don't divide by zero, and handle nan/inf.
    normal_scale_val = np_api.logical_and(np_api.isfinite(scale_pow2), scale_pow2 != 0)
    scale_renorm = np_api.where(normal_scale_val, scale_pow2, minval)
    mantissa = val / scale_renorm
    return scale_pow2.astype(scale_dtype), mantissa.astype(vin.dtype)


class Pow2RoundMode(IntEnum):
    """Power-of-two supported rounded mode."""

    NONE = 0
    DOWN = 1
    UP = 2
    STOCHASTIC = 3


pow2_decompose_p = core.Primitive("pow2_decompose")
"""`pow2_decompose` pow2 decompose JAX primitive.
"""


def pow2_decompose(
    vin: Array, scale_dtype: Optional[DTypeLike] = None, mode: Pow2RoundMode = Pow2RoundMode.DOWN
) -> Tuple[Array, Array]:
    """Power-2 decompose, i.e. vin = s * vout where s is a power-of 2 scaling.

    Args:
        vin: Input array.
        scale_dtype: Scale dtype to use.
        mode: Pow2 rounding.
    Returns:
        (scale, vout) such that vin = scale * vout
    """
    scale_dtype = np.dtype(scale_dtype or vin.dtype)
    # A couple of checks on dtypes.
    assert np.issubdtype(vin.dtype, np.floating)
    assert np.issubdtype(scale_dtype, np.floating)
    if scale_dtype == np.float16:
        logging.warning("`pow2_decompose` does not support FP16 sub-normals when using FP16 scale dtype.")
    out = pow2_decompose_p.bind(vin, scale_dtype=scale_dtype, mode=mode)
    return out


def pow2_decompose_eager_impl(
    vin: Array, scale_dtype: Optional[DTypeLike] = None, mode: Pow2RoundMode = Pow2RoundMode.DOWN
) -> Tuple[Array, Array]:
    """Eager mode implementation, on JAX/Numpy arrays."""
    if mode == Pow2RoundMode.DOWN:
        return pow2_decompose_round_down_impl(vin, scale_dtype)
    raise NotImplementedError(f"Unsupported power-of-2 rounding mode '{mode}'.")


def pow2_decompose_abstract_eval(
    vin: core.ShapedArray, scale_dtype: Optional[DTypeLike] = None, mode: Pow2RoundMode = Pow2RoundMode.DOWN
) -> Tuple[core.ShapedArray, core.ShapedArray]:
    scale_dtype = scale_dtype or vin.dtype
    sout = core.ShapedArray(vin.shape, dtype=scale_dtype)
    return (sout, vin)


def pow2_decompose_mlir_lowering(
    ctx: LoweringRuleContext, *args: Union[ir.Value, Sequence[ir.Value]], **params: Dict[str, Any]
) -> Sequence[Union[ir.Value, Sequence[ir.Value]]]:
    scale_dtype = params["scale_dtype"]
    mode = params["mode"]
    pow2_decompose_fn = partial(pow2_decompose_eager_impl, scale_dtype=scale_dtype, mode=mode)
    outputs = mlir.lower_fun(pow2_decompose_fn, multiple_results=True)(ctx, *args)
    return outputs


# Register as standard JAX primitive
pow2_decompose_p.multiple_results = True
pow2_decompose_p.def_abstract_eval(pow2_decompose_abstract_eval)
pow2_decompose_p.def_impl(pow2_decompose_eager_impl)
# Default lowering on GPU, TPU, ...
mlir.register_lowering(pow2_decompose_p, pow2_decompose_mlir_lowering)


def pow2_round_down(val: Array) -> Array:
    """Round down to the closest power of 2."""
    # Keep only the scale component of `pow2_decompose`
    pow2_val, _ = pow2_decompose(val, scale_dtype=val.dtype, mode=Pow2RoundMode.DOWN)
    return pow2_val


def pow2_round_up(val: Array) -> Array:
    """Round up to the closest power of 2.
    NOTE: may overflow to inf.
    """
    # FIXME: rounding when already a power of 2.
    # Should do additional masking to check that.
    pow2_val = pow2_round_down(val) * np.array(2, dtype=val.dtype)
    return pow2_val


def pow2_round(val: Array, mode: Pow2RoundMode = Pow2RoundMode.DOWN) -> Array:
    """Power-of-two rounding."""
    if mode == Pow2RoundMode.NONE:
        return val
    elif mode == Pow2RoundMode.DOWN:
        return pow2_round_down(val)
    elif mode == Pow2RoundMode.UP:
        return pow2_round_up(val)
    raise NotImplementedError(f"Unsupported power-of-2 rounding mode '{mode}'.")


def get_mantissa(val: Array) -> Array:
    """Extract the mantissa of an array, masking the exponent.

    Similar to `numpy.frexp`, but with implicit bit to be consistent with
    `pow2_round_down`.
    """
    _, mantissa = pow2_decompose(val, scale_dtype=val.dtype, mode=Pow2RoundMode.DOWN)
    return mantissa



# Copyright (c) 2023 Graphcore Ltd. All rights reserved.
from typing import Any, Callable, Dict

from jax import tree_util
from jax._src.debugging import debug_callback as debug_callback_orig
from jax._src.debugging import debug_callback_p

from .interpreters import ScaledArray, register_scaled_op


def get_debug_callback_effect(ordered: bool) -> Any:
    """Backward compatible effect factory method."""
    try:
        from jax._src.debugging import debug_effect, ordered_debug_effect

        return ordered_debug_effect if ordered else debug_effect
    except ImportError:
        from jax._src.debugging import DebugEffect

        return DebugEffect.ORDERED_PRINT if ordered else DebugEffect.PRINT


def debug_callback(callback: Callable[..., Any], *args: Any, ordered: bool = False, **kwargs: Any) -> None:
    # We need our custom version of `debug_callback` to deal with
    # changing JAX pytrees.
    # FIXME: probably patch `debug_callback` in JAX.
    flat_args, in_tree = tree_util.tree_flatten((args, kwargs))
    effect = get_debug_callback_effect(ordered)

    def _flat_callback(*flat_args):
        args, kwargs = tree_util.tree_unflatten(in_tree, flat_args)
        callback(*args, **kwargs)
        return []

    # Storing in original PyTree and callback function.
    # Allowing custom interpreters to retrieve and modify this information.
    _flat_callback.__callback_fn = callback  # type:ignore
    _flat_callback.__callback_in_tree = in_tree  # type:ignore
    debug_callback_p.bind(*flat_args, callback=_flat_callback, effect=effect)


debug_callback.__doc__ = debug_callback_orig.__doc__


def scaled_debug_callback(*args: ScaledArray, **params: Dict[str, Any]) -> Any:
    """Scaled `debug_callback`: properly forwarding ScaledArrays
    to host callback.
    """
    flat_callback_fn = params["callback"]
    if not hasattr(flat_callback_fn, "__callback_fn"):
        raise NotImplementedError("Please use `jsa.debug_callback` function instead of original JAX function.")
    callback_fn = flat_callback_fn.__callback_fn
    in_pytree = flat_callback_fn.__callback_in_tree  # type:ignore
    # Re-build original input, with scaled arrays.
    scaled_args, scaled_kwargs = tree_util.tree_unflatten(in_pytree, args)
    # Re-build ordered boolean, in a backward compatible way.
    ordered = "ordered" in str(params["effect"]).lower()
    debug_callback(callback_fn, *scaled_args, ordered=ordered, **scaled_kwargs)
    return []


register_scaled_op(debug_callback_p, scaled_debug_callback)



# Copyright (c) 2024 Graphcore Ltd. All rights reserved.
from functools import partial

import jax


@partial(jax.custom_vjp, nondiff_argnums=(0,))
def map_on_forward(f, arg):
    """Map a function on a forward pass only. No-op/identity on backward pass."""
    return f(arg)


def map_on_forward_fwd(f, arg):
    return arg, None


def map_on_forward_bwd(f, _, grad):
    return (grad,)


map_on_forward.defvjp(map_on_forward_fwd, map_on_forward_bwd)


@partial(jax.custom_vjp, nondiff_argnums=(0,))
def map_on_backward(f, arg):
    """Map a function on the gradient/backward pass. No-op/identity on forward."""
    return arg


def map_on_backward_fwd(f, arg):
    return arg, None


def map_on_backward_bwd(f, _, grad):
    return (f(grad),)


map_on_backward.defvjp(map_on_backward_fwd, map_on_backward_bwd)



# Copyright (c) 2023 Graphcore Ltd. All rights reserved.
import chex
import jax.numpy as jnp
import numpy as np
import numpy.testing as npt
from absl.testing import parameterized
from jax.core import ShapedArray

from jax_scalify.core import (
    Array,
    ScaledArray,
    as_scaled_array,
    asarray,
    get_scale_dtype,
    is_scaled_leaf,
    is_static_anyscale,
    is_static_one_scalar,
    is_static_zero,
    make_scaled_scalar,
    pow2_round_down,
    scaled_array,
)


class ScaledArrayDataclassTests(chex.TestCase):
    @parameterized.parameters(
        {"npapi": np},
        {"npapi": jnp},
    )
    def test__scaled_array__init__multi_numpy_backend(self, npapi):
        sarr = ScaledArray(data=npapi.array([1.0, 2.0], dtype=np.float32), scale=npapi.array(1))
        assert isinstance(sarr.data, npapi.ndarray)
        assert isinstance(sarr.scale, npapi.ndarray)
        assert sarr.scale.shape == ()

    @parameterized.parameters(
        {"npapi": np},
        {"npapi": jnp},
    )
    def test__scaled_array__factory_method__multi_numpy_backend(self, npapi):
        sarr = scaled_array(data=[1.0, 2.0], scale=3, dtype=np.float16, npapi=npapi)
        assert isinstance(sarr, ScaledArray)
        assert isinstance(sarr.data, npapi.ndarray)
        assert isinstance(sarr.scale, npapi.ndarray)
        assert sarr.data.dtype == ShapedArray((2,), np.float16)
        assert sarr.scale.shape == ()
        npt.assert_array_almost_equal(sarr, [3, 6])

    @parameterized.parameters(
        {"dtype": np.float16},
        {"dtype": np.float32},
    )
    def test__scaled_array__factory_method__multi_dtypes(self, dtype):
        sarr = scaled_array(data=[1.0, 2.0], scale=3.0, dtype=dtype, npapi=jnp)
        assert isinstance(sarr, ScaledArray)
        assert isinstance(sarr.data, jnp.ndarray)
        assert isinstance(sarr.scale, jnp.ndarray)
        assert sarr.data.dtype == ShapedArray((2,), dtype=dtype)
        # Default scale dtype is always FP32.
        # TODO: what happens when integer scale is passed?
        assert sarr.scale.dtype == np.float32
        assert sarr.scale.shape == ()
        npt.assert_array_almost_equal(sarr, [3, 6])

    @parameterized.parameters(
        {"npapi": np},
        {"npapi": jnp},
    )
    def test__scaled_array__basic_properties(self, npapi):
        sarr = ScaledArray(data=npapi.array([1.0, 2.0], dtype=np.float32), scale=npapi.array(1))
        assert sarr.dtype == np.float32
        assert sarr.shape == (2,)
        assert sarr.aval == ShapedArray((2,), np.float32)

    @parameterized.parameters(
        {"npapi": np},
        {"npapi": jnp},
    )
    def test__scaled_array__to_array__multi_numpy_backend(self, npapi):
        sarr = scaled_array(data=[1.0, 2.0], scale=3, dtype=np.float16, npapi=npapi)
        # No dtype specified.
        out = sarr.to_array()
        assert isinstance(out, npapi.ndarray)
        assert out.dtype == sarr.dtype
        npt.assert_array_equal(out, sarr.data * sarr.scale)
        # Custom float dtype.
        out = sarr.to_array(dtype=np.float32)
        assert isinstance(out, npapi.ndarray)
        assert out.dtype == np.float32
        npt.assert_array_equal(out, sarr.data * sarr.scale)
        # Custom int dtype.
        out = sarr.to_array(dtype=np.int8)
        assert isinstance(out, npapi.ndarray)
        assert out.dtype == np.int8
        npt.assert_array_equal(out, sarr.data * sarr.scale)

    @parameterized.parameters(
        {"npapi": np},
        {"npapi": jnp},
    )
    def test__scaled_array__numpy_array_interface(self, npapi):
        sarr = ScaledArray(data=npapi.array([1.0, 2.0], dtype=np.float32), scale=npapi.array(3))
        out = np.asarray(sarr)
        assert isinstance(out, np.ndarray)
        npt.assert_array_equal(out, sarr.data * sarr.scale)

    @parameterized.parameters(
        {"npapi": np},
        {"npapi": jnp},
    )
    def test__scaled_array__astype(self, npapi):
        arr_in = ScaledArray(data=npapi.array([1.0, 2.0], dtype=np.float16), scale=npapi.array(1, dtype=np.int32))
        arr_out = arr_in.astype(np.float32)
        assert arr_out.dtype == np.float32
        assert arr_out.scale.dtype == arr_in.scale.dtype

    @parameterized.parameters(
        {"val": 0.25},
    )
    def test__make_scaled_scalar__float_input(self, val):
        scaled_val = make_scaled_scalar(val)
        assert isinstance(scaled_val, ScaledArray)
        assert scaled_val.shape == ()
        assert scaled_val.data.dtype == np.float32
        assert scaled_val.scale.dtype == np.float32
        npt.assert_equal(np.asarray(scaled_val), val)
        assert isinstance(scaled_val.data, (np.ndarray, np.number))
        assert isinstance(scaled_val.scale, (np.ndarray, np.number))

    @parameterized.parameters(
        {"val": np.float16(0)},
        {"val": np.float32(0)},
        # {"val": np.float64(0)}, FIXME!
    )
    def test__make_scaled_scalar__zero_scalar_input(self, val):
        scaled_val = make_scaled_scalar(val)
        assert isinstance(scaled_val, ScaledArray)
        assert scaled_val.shape == ()
        assert scaled_val.dtype == val.dtype

    def test__make_scaled_scalar__optional_scale_dtype(self):
        val = np.float16(0.25)
        scaled_val = make_scaled_scalar(val, scale_dtype=np.float32)
        assert isinstance(scaled_val, ScaledArray)
        assert scaled_val.dtype == val.dtype
        assert scaled_val.scale.dtype == np.float32
        npt.assert_equal(np.asarray(scaled_val), val)

    @parameterized.parameters(
        {"val": np.finfo(np.float16).smallest_normal},
        {"val": np.finfo(np.float16).smallest_subnormal},
        {"val": np.float16(3.123283386230469e-05)},
    )
    def test__make_scaled_scalar__fp16_subnormal_support(self, val):
        # Use FP32 scale dtype, to have enough range.
        # NOTE: failing in FP16!
        scaled_val = make_scaled_scalar(val, scale_dtype=np.float32)
        # No loss of information when converting everything to FP32.
        npt.assert_equal(np.asarray(scaled_val, dtype=np.float32), np.float32(val))

    @parameterized.parameters(
        {"val": np.array(1.0)},
        {"val": np.float32(-0.5)},
        {"val": np.float16(1.25)},
        {"val": np.float32(-65504)},
        # Testing JAX arrays too.
        {"val": jnp.float32(0.5)},
        {"val": jnp.float16(1.25)},
    )
    def test__make_scaled_scalar__proper_split_data_scale(self, val):
        scaled_val = make_scaled_scalar(val)
        assert isinstance(scaled_val, ScaledArray)
        assert scaled_val.shape == ()
        assert scaled_val.data.dtype == val.dtype
        assert scaled_val.scale.dtype == val.dtype
        npt.assert_equal(np.asarray(scaled_val), val)
        npt.assert_equal(scaled_val.scale, pow2_round_down(val))
        npt.assert_array_less(0, scaled_val.scale)
        # Make sure we are not doing implicit conversion to JAX arrays.
        if isinstance(val, Array):
            assert isinstance(scaled_val.data, Array)
            assert isinstance(scaled_val.scale, Array)
        else:
            assert isinstance(scaled_val.data, (np.ndarray, np.number))
            assert isinstance(scaled_val.scale, (np.ndarray, np.number))

    def test__is_scaled_leaf__consistent_with_jax(self):
        assert is_scaled_leaf(8)
        assert is_scaled_leaf(2.0)
        assert is_scaled_leaf(np.int32(2))
        assert is_scaled_leaf(np.float32(2))
        assert is_scaled_leaf(np.array(3))
        assert is_scaled_leaf(np.array([3]))
        assert is_scaled_leaf(jnp.array([3]))
        assert is_scaled_leaf(scaled_array(data=[1.0, 2.0], scale=3, dtype=np.float16))

    @parameterized.parameters(
        {"data": np.array([1, 2.0])},
        {"data": jnp.array([1, 2.0])},
    )
    def test__as_scaled_array__unchanged_dtype(self, data):
        output = as_scaled_array(data)
        assert isinstance(output, ScaledArray)
        assert isinstance(output.data, type(data))
        assert output.dtype in {np.dtype(np.float32), np.dtype(np.float64)}
        npt.assert_array_almost_equal(output, data)
        npt.assert_array_equal(output.scale, np.array(1, dtype=data.dtype))
        # unvariant when calling a second time.
        assert as_scaled_array(output) is output

    @parameterized.parameters(
        {"data": 2.1},
        {"data": np.float64(2.0)},
        {"data": jnp.float32(2.0)},
    )
    def test__as_scaled_array__float_scalar(self, data):
        output = as_scaled_array(data)
        assert isinstance(output, ScaledArray)
        assert output.data.dtype == output.scale.dtype
        # NOTE: for scalars, data always in [1, 2)
        npt.assert_almost_equal(np.asarray(output), data)

    @parameterized.parameters(
        {"data": jnp.float32(3.0)},
    )
    def test__as_scaled_array__optional_scalar(self, data):
        scale = np.float16(2)
        output = as_scaled_array(data, scale=scale)
        assert isinstance(output, ScaledArray)
        assert output.scale.dtype == np.float16
        npt.assert_array_equal(output.scale, np.array(2, dtype=np.float16))
        npt.assert_array_almost_equal(output, data)

    @parameterized.parameters(
        {"data": False},
        {"data": 2},
        {"data": np.array([1, 2])},
        {"data": jnp.array([1, 2])},
    )
    def test__as_scaled_array__unscaled_bool_int_output(self, data):
        output = as_scaled_array(data)
        assert output is data

    @chex.variants(with_jit=True, without_jit=True)
    def test__as_scaled_array__complex_pytree(self):
        input = {"x": jnp.array([1, 2]), "y": jnp.array([1.0, 2]), "z": as_scaled_array(jnp.array([1.0, 2]))}
        output = self.variant(as_scaled_array)(input, scale=np.float32(2))
        assert isinstance(output, dict)
        assert len(output) == 3

        npt.assert_array_equal(output["x"], input["x"])
        npt.assert_array_equal(output["y"], input["y"])
        npt.assert_array_equal(output["z"], input["z"])
        npt.assert_almost_equal(output["y"].scale, 2)

        if "without_jit" in self.variant.__name__:
            assert output["x"] is input["x"]
            assert output["z"] is input["z"]

    @chex.variants(with_jit=True, without_jit=True)
    @parameterized.parameters(
        {"data": np.int32(3)},
        {"data": np.array(2, dtype=np.int32)},
        {"data": np.array([1, 2], dtype=np.int32)},
        {"data": np.array([1, 2.0], dtype=np.float32)},
        {"data": jnp.array([1, 2])},
    )
    def test__asarray__unchanged_dtype(self, data):
        output = self.variant(asarray)(data)
        assert output.dtype == data.dtype
        npt.assert_array_almost_equal(output, data)

    @parameterized.parameters(
        {"data": np.array([1, 2])},
        {"data": jnp.array([1, 2])},
        {"data": scaled_array(data=[1.0, 2.0], scale=3, dtype=np.float32)},
    )
    def test__asarray__changed_dtype(self, data):
        output = asarray(data, dtype=np.float16)
        assert output.dtype == np.float16
        npt.assert_array_almost_equal(output, data)

    def test__asarray__complex_pytree(self):
        input = {"x": jnp.array([1.0, 2]), "y": scaled_array(jnp.array([3, 4.0]), jnp.array(0.5))}
        output = asarray(input)
        assert isinstance(output, dict)
        assert len(output) == 2
        assert all([isinstance(v, Array) for v in output.values()])
        npt.assert_array_almost_equal(output["x"], input["x"])
        npt.assert_array_almost_equal(output["y"], input["y"])

    @parameterized.parameters(
        {"val": 0, "result": True},
        {"val": 0.0, "result": True},
        {"val": np.int32(0), "result": True},
        {"val": np.float16(0), "result": True},
        {"val": np.array([1, 2]), "result": False},
        {"val": np.array([0, 0.0]), "result": True},
        {"val": jnp.array([0, 0.0]), "result": False},
        {"val": ScaledArray(np.array([0, 0.0]), jnp.array(2.0)), "result": True},
        {"val": ScaledArray(jnp.array([3, 4.0]), np.array(0.0)), "result": True},
        {"val": ScaledArray(jnp.array([3, 4.0]), jnp.array(0.0)), "result": False},
    )
    def test__is_static_zero__proper_all_result(self, val, result):
        all_zero = np.all(is_static_zero(val))
        assert all_zero == result

    @parameterized.parameters(
        {"val": 0, "result": True},
        {"val": 0.0, "result": True},
        {"val": np.inf, "result": True},
        {"val": np.int32(0), "result": True},
        {"val": np.float16(0), "result": True},
        {"val": np.float16(-np.inf), "result": True},
        {"val": np.array([1, 2]), "result": False},
        {"val": np.array([0, 0.0, -np.inf, np.inf]), "result": True},
        {"val": jnp.array([0, 0.0, np.inf]), "result": False},
        {"val": ScaledArray(np.array([0, 0.0]), jnp.array(2.0)), "result": True},
        {"val": ScaledArray(jnp.array([3, 4.0]), np.array(0.0)), "result": True},
        {"val": ScaledArray(jnp.array([3, 4.0]), np.array(np.inf)), "result": True},
        {"val": ScaledArray(jnp.array([3, 4.0]), jnp.array(0.0)), "result": False},
    )
    def test__is_static_anyscale__proper_all_result(self, val, result):
        all_zero = np.all(is_static_anyscale(val))
        assert all_zero == result

    @parameterized.parameters(
        {"val": 0, "result": False},
        {"val": 1, "result": True},
        {"val": 1.0, "result": True},
        {"val": np.int32(1), "result": True},
        {"val": np.float16(1), "result": True},
        {"val": np.float32(1), "result": True},
        {"val": np.array([1]), "result": True},
        {"val": np.array([1]), "result": True},
        {"val": np.array([1, 1]), "result": False},
        {"val": jnp.array([1, 1]), "result": False},
        {"val": ScaledArray(np.array([1]), np.array(1.0)), "result": True},
        {"val": ScaledArray(np.array(1.0), np.array(1)), "result": True},
        {"val": ScaledArray(np.array([1]), np.array(2.0)), "result": False},
    )
    def test__is_static_one_scalar__proper_result(self, val, result):
        r = is_static_one_scalar(val)
        assert isinstance(r, (bool, np.bool_))
        assert r == result
        # Should still be valid when we use `asarray` as well.
        r = is_static_one_scalar(asarray(val))
        assert isinstance(r, (bool, np.bool_))
        assert r == result

    @parameterized.parameters(
        {"val": np.float32(2), "result": np.float32},
        {"val": ScaledArray(np.array(2.0), np.float16(3)), "result": np.float16},
    )
    def test__get_scale_dtype__proper_dtype(self, val, result):
        sdtype = get_scale_dtype(val)
        assert sdtype == result



# Copyright (c) 2023 Graphcore Ltd. All rights reserved.
from functools import partial

import jax
import numpy as np

from jax_scalify.core import ScaledArray, pow2_round, pow2_round_down
from jax_scalify.lax import get_data_scale, rebalance

from .utils import map_on_backward, map_on_forward


def dynamic_rescale_max_base(arr: ScaledArray) -> ScaledArray:
    """Dynamic rescaling of a ScaledArray, using abs-max."""
    # Similarly to ML norms => need some epsilon for training stability!
    eps = pow2_round_down(np.float32(1e-4))

    data, scale = get_data_scale(arr)
    data_sq = jax.lax.abs(data)
    axes = tuple(range(data.ndim))
    # Get MAX norm + pow2 rounding.
    norm = jax.lax.reduce_max_p.bind(data_sq, axes=axes)
    norm = jax.lax.max(pow2_round(norm).astype(scale.dtype), eps.astype(scale.dtype))
    # Rebalancing based on norm.
    return rebalance(arr, norm)


def dynamic_rescale_l1_base(arr: ScaledArray) -> ScaledArray:
    """Dynamic rescaling of a ScaledArray, using L1 norm.

    NOTE: by default, computing L1 norm in FP32.
    """
    # Similarly to ML norms => need some epsilon for training stability!
    norm_dtype = np.float32
    eps = pow2_round_down(norm_dtype(1e-4))

    data, scale = get_data_scale(arr)
    data_sq = jax.lax.abs(data.astype(np.float32))
    axes = tuple(range(data.ndim))
    # Get L1 norm + pow2 rounding.
    norm = jax.lax.reduce_sum_p.bind(data_sq, axes=axes) / data.size
    norm = jax.lax.max(pow2_round(norm), eps).astype(scale.dtype)
    # Rebalancing based on norm.
    return rebalance(arr, norm)


def dynamic_rescale_l2_base(arr: ScaledArray) -> ScaledArray:
    """Dynamic rescaling of a ScaledArray, using L2 norm.

    NOTE: by default, computing L2 norm in FP32.
    """
    # Similarly to ML norms => need some epsilon for training stability!
    norm_dtype = np.float32
    eps = pow2_round_down(norm_dtype(1e-4))

    data, scale = get_data_scale(arr)
    data_sq = jax.lax.integer_pow(data.astype(norm_dtype), 2)
    axes = tuple(range(data.ndim))
    # Get L2 norm + pow2 rounding.
    norm = jax.lax.sqrt(jax.lax.reduce_sum_p.bind(data_sq, axes=axes) / data.size)
    # Make sure we don't "underflow" too much on the norm.
    norm = jax.lax.max(pow2_round(norm), eps).astype(scale.dtype)
    # Rebalancing based on norm.
    return rebalance(arr, norm)


# Dynamic rescale on fwd arrays.
dynamic_rescale_max = partial(map_on_forward, dynamic_rescale_max_base)
dynamic_rescale_l1 = partial(map_on_forward, dynamic_rescale_l1_base)
dynamic_rescale_l2 = partial(map_on_forward, dynamic_rescale_l2_base)

# Dynamic rescale on gradients.
dynamic_rescale_max_grad = partial(map_on_backward, dynamic_rescale_max_base)
dynamic_rescale_l1_grad = partial(map_on_backward, dynamic_rescale_l1_base)
dynamic_rescale_l2_grad = partial(map_on_backward, dynamic_rescale_l2_base)


# Backward compatibility. DEPRECATED, WILL BE REMOVED!
fn_fwd_identity_bwd = map_on_forward
fn_bwd_identity_fwd = map_on_backward



# Copyright 2018 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Modified by Graphcore Ltd 2024.

"""A basic CIFAR10 example using Numpy and JAX.

CIFAR10 training using MLP network + raw SGD optimizer.
"""
import time

import dataset_cifar10
import jax
import jax.numpy as jnp
import numpy as np
import numpy.random as npr
from jax import grad, jit, lax

import jax_scalify as jsa


def logsumexp(a, axis=None, keepdims=False):
    dims = (axis,)
    amax = jnp.max(a, axis=dims, keepdims=keepdims)
    # FIXME: not proper scale propagation, introducing NaNs
    # amax = lax.stop_gradient(lax.select(jnp.isfinite(amax), amax, lax.full_like(amax, 0)))
    amax = lax.stop_gradient(amax)
    out = lax.sub(a, amax)
    out = lax.exp(out)
    out = lax.add(lax.log(jnp.sum(out, axis=dims, keepdims=keepdims)), amax)
    return out


def init_random_params(scale, layer_sizes, rng=npr.RandomState(0)):
    return [(scale * rng.randn(m, n), scale * rng.randn(n)) for m, n, in zip(layer_sizes[:-1], layer_sizes[1:])]


def print_mean_std(name, v):
    data, scale = jsa.lax.get_data_scale(v)
    # Always use np.float32, to avoid floating errors in descaling + stats.
    v = jsa.asarray(data, dtype=np.float32)
    m, s = np.mean(v), np.std(v)
    # print(data)
    print(f"{name}: MEAN({m:.4f}) / STD({s:.4f}) / SCALE({scale:.4f})")


def predict(params, inputs):
    activations = inputs
    for w, b in params[:-1]:
        # Matmul + relu
        outputs = jnp.dot(activations, w) + b
        activations = jnp.maximum(outputs, 0)

    final_w, final_b = params[-1]
    logits = jnp.dot(activations, final_w) + final_b

    # Dynamic rescaling of the gradient, as logits gradient not properly scaled.
    logits = jsa.ops.dynamic_rescale_l2_grad(logits)
    output = logits - logsumexp(logits, axis=1, keepdims=True)

    return output


def loss(params, batch):
    inputs, targets = batch
    preds = predict(params, inputs)
    return -jnp.mean(jnp.sum(preds * targets, axis=1))


def accuracy(params, batch):
    inputs, targets = batch
    target_class = jnp.argmax(targets, axis=1)
    predicted_class = jnp.argmax(predict(params, inputs), axis=1)
    return jnp.mean(predicted_class == target_class)


if __name__ == "__main__":
    width = 2048
    lr = 1e-4
    use_scalify = True
    scalify = jsa.scalify if use_scalify else lambda f: f

    layer_sizes = [3072, width, width, 10]
    param_scale = 1.0

    step_size = lr
    num_epochs = 10
    batch_size = 128
    training_dtype = np.float16
    scale_dtype = np.float32

    train_images, train_labels, test_images, test_labels = dataset_cifar10.cifar()
    num_train = train_images.shape[0]
    num_complete_batches, leftover = divmod(num_train, batch_size)
    num_batches = num_complete_batches + bool(leftover)

    def data_stream():
        rng = npr.RandomState(0)
        while True:
            perm = rng.permutation(num_train)
            for i in range(num_batches):
                batch_idx = perm[i * batch_size : (i + 1) * batch_size]
                yield train_images[batch_idx], train_labels[batch_idx]

    batches = data_stream()
    params = init_random_params(param_scale, layer_sizes)
    # Transform parameters to `ScaledArray` and proper dtype.
    if use_scalify:
        params = jsa.as_scaled_array(params, scale=scale_dtype(param_scale))
    params = jax.tree_util.tree_map(lambda v: v.astype(training_dtype), params, is_leaf=jsa.core.is_scaled_leaf)

    @jit
    @scalify
    def update(params, batch):
        grads = grad(loss)(params, batch)
        return [(w - step_size * dw, b - step_size * db) for (w, b), (dw, db) in zip(params, grads)]

    for epoch in range(num_epochs):
        start_time = time.time()
        for _ in range(num_batches):
            batch = next(batches)
            # Scaled micro-batch + training dtype cast.
            if use_scalify:
                batch = jsa.as_scaled_array(batch, scale=scale_dtype(param_scale))
            batch = jax.tree_util.tree_map(lambda v: v.astype(training_dtype), batch, is_leaf=jsa.core.is_scaled_leaf)

            with jsa.ScalifyConfig(rounding_mode=jsa.Pow2RoundMode.DOWN, scale_dtype=scale_dtype):
                params = update(params, batch)

        epoch_time = time.time() - start_time

        # Evaluation in float32, for consistency.
        raw_params = jsa.asarray(params, dtype=np.float32)
        train_acc = accuracy(raw_params, (train_images, train_labels))
        test_acc = accuracy(raw_params, (test_images, test_labels))
        print(f"Epoch {epoch} in {epoch_time:0.2f} sec")
        print(f"Training set accuracy {train_acc:0.5f}")
        print(f"Test set accuracy {test_acc:0.5f}")



# Copyright (c) 2023 Graphcore Ltd. All rights reserved.
import chex
import jax
import jax.lax
import jax.numpy as jnp
import numpy as np
import numpy.testing as npt
from absl.testing import parameterized
from numpy.typing import NDArray

from jax_scalify.core import (
    Array,
    Pow2RoundMode,
    ScaledArray,
    ScalifyConfig,
    asarray,
    get_scalify_config,
    is_scaled_leaf,
    register_scaled_op,
    scaled_array,
    scalify,
)
from jax_scalify.core.interpreters import ScalifyTracerArray


class ScalifyTracerArrayTests(chex.TestCase):
    @parameterized.parameters(
        {"arr": True},
        {"arr": 2},
        {"arr": 3.0},
    )
    def test__scalify_tracer_array__init__from_python_value(self, arr):
        tracer_arr = ScalifyTracerArray(arr)
        assert tracer_arr.array == arr
        assert not tracer_arr.is_scaled_array
        assert tracer_arr.is_broadcasted_scalar == (tracer_arr.size == 1)
        assert not tracer_arr.is_broadcasted_zero
        assert tracer_arr.to_array() is tracer_arr.array

    @parameterized.parameters(
        {"arr": np.float32(2)},
        {"arr": np.array([1, 2])},
        {"arr": jnp.array([3, 4])},
    )
    def test__scalify_tracer_array__init__from_normal_array(self, arr):
        tracer_arr = ScalifyTracerArray(arr)
        assert tracer_arr.array is arr
        assert not tracer_arr.is_scaled_array
        assert tracer_arr.is_broadcasted_scalar == (tracer_arr.size == 1)
        assert not tracer_arr.is_broadcasted_zero
        assert tracer_arr.to_array() is tracer_arr.array
        # Basic properties.
        assert tracer_arr.shape == arr.shape
        assert tracer_arr.size == arr.size

    @parameterized.parameters(
        {"arr": np.float32(2), "expected_is_zero": False},
        {"arr": np.float32(0), "expected_is_zero": True},
        {"arr": np.array([0, 0]), "expected_is_zero": True},
        {"arr": np.array([0.0, 0.0]), "expected_is_zero": True},
        {"arr": scaled_array([1, 2.0], 0.0, npapi=np), "expected_is_zero": True},
        {"arr": scaled_array([0, 0.0], 1.0, npapi=np), "expected_is_zero": True},
        {"arr": jnp.array([0, 0]), "expected_is_zero": False},
    )
    def test__scalify_tracer_array__init__zero_broadcasted_array(self, arr, expected_is_zero):
        tracer_arr = ScalifyTracerArray(arr)
        assert tracer_arr.is_broadcasted_zero == expected_is_zero
        # Scaled array conversion => scale should be zero.
        scaled_arr = tracer_arr.to_scaled_array()
        if tracer_arr.is_broadcasted_zero and isinstance(scaled_arr, ScaledArray):
            assert scaled_arr.scale == 0

    @parameterized.parameters({"arr": scaled_array([1, 2], 3.0)})
    def test__scalify_tracer_array__init__from_scaled_array(self, arr):
        tracer_arr = ScalifyTracerArray(arr)
        assert tracer_arr.array is arr
        assert tracer_arr.is_scaled_array
        assert tracer_arr.to_scaled_array() is tracer_arr.array
        assert not tracer_arr.is_broadcasted_zero

    def test__scalify_tracer_array__init__is_broadcasted_scalar_kwarg(self):
        arr = scaled_array([1, 2], 3.0)
        assert ScalifyTracerArray(arr, is_broadcasted_scalar=True).is_broadcasted_scalar
        assert not ScalifyTracerArray(arr, is_broadcasted_scalar=False).is_broadcasted_scalar

    def test__scalify_tracer_array__init__is_broadcasted_zero_kwarg(self):
        arr = scaled_array([0, 1], 3.0)
        # NOTE: explicitly passing the argument, not checking the data!
        assert ScalifyTracerArray(arr, is_broadcasted_zero=True).is_broadcasted_scalar
        assert ScalifyTracerArray(arr, is_broadcasted_zero=True).is_broadcasted_zero
        assert not ScalifyTracerArray(arr, is_broadcasted_zero=False).is_broadcasted_scalar
        assert not ScalifyTracerArray(arr, is_broadcasted_zero=False).is_broadcasted_zero

    def test__scalify_tracer_array__flatten__proper_pytree(self):
        arr = scaled_array([1, 2], 3.0)
        tracer_arr_in = ScalifyTracerArray(arr, is_broadcasted_scalar=True, is_broadcasted_zero=True)
        # Proper round trip!
        flat_arrays, pytree = jax.tree_util.tree_flatten(tracer_arr_in)
        tracer_arr_out = jax.tree_util.tree_unflatten(pytree, flat_arrays)

        assert isinstance(tracer_arr_out, ScalifyTracerArray)
        assert tracer_arr_out.is_broadcasted_scalar == tracer_arr_in.is_broadcasted_scalar
        assert tracer_arr_out.is_broadcasted_zero == tracer_arr_in.is_broadcasted_zero
        npt.assert_array_equal(np.asarray(tracer_arr_out.array), np.asarray(tracer_arr_in.array))

    @parameterized.parameters(
        {"input": 3.0},
        {"input": np.float32(3.0)},
        {"input": np.array(3.0)},
        {"input": jnp.array(3.0)},
    )
    def test__scalify_tracer_array__to_scaled_array__scalar_input(self, input):
        scaled_val = ScalifyTracerArray(input).to_scaled_array()
        assert isinstance(scaled_val, ScaledArray)
        assert scaled_val.data.dtype == scaled_val.scale.dtype
        # NOTE: scale is a power-of-two.
        npt.assert_almost_equal(np.asarray(scaled_val), input)

    @parameterized.parameters(
        {"input": np.array(3)},
        {"input": jnp.array(3)},
        {"input": np.int32(2)},
    )
    def test__scalify_tracer_array__to_scaled_array__not_promoted_input(self, input):
        out = ScalifyTracerArray(input).to_scaled_array()
        assert out is input

    def test__scalify_tracer_array__to_scaled_array__broadcasted_scalar_input(self):
        data: NDArray[np.float16] = np.array([5, 5], dtype=np.float16)
        scaled_out = ScalifyTracerArray(data, is_broadcasted_scalar=True).to_scaled_array(scale_dtype=np.float32)

        assert isinstance(scaled_out, ScaledArray)
        assert scaled_out.dtype == data.dtype
        assert scaled_out.scale.dtype == np.float32
        npt.assert_almost_equal(scaled_out.scale, 4)
        npt.assert_array_equal(np.asarray(scaled_out), data)


class ScalifyInterpreterTests(chex.TestCase):
    def test__register_scaled_op__error_if_already_registered(self):
        with self.assertRaises(KeyError):
            register_scaled_op(jax.lax.mul_p, lambda a, _: a)

    @chex.variants(with_jit=True, without_jit=True)
    def test__scalify_interpreter__normal_jax_mode(self):
        def func(x):
            return x * 2

        func = self.variant(scalify(func))
        data: NDArray[np.float32] = np.array([1, 2], dtype=np.float32)
        out = func(data)
        # Proper behaviour!
        assert isinstance(out, Array)
        npt.assert_array_equal(out, [2, 4])
        # Check jaxpr.
        jaxpr = jax.make_jaxpr(func)(data).jaxpr
        assert len(jaxpr.invars) == 1
        assert len(jaxpr.outvars) == 1
        assert len(jaxpr.eqns) == 1

    def test__scalify_interpreter__without_jit__proper_jaxpr_signature(self):
        def func(x):
            return x * 2

        scaled_func = scalify(func)
        scaled_input = scaled_array([1.0, 2.0], 3, dtype=np.float32)
        jaxpr = jax.make_jaxpr(scaled_func)(scaled_input).jaxpr
        # Need 4 equations: 1 pow2_decompose + 2 mul + 1 cast.
        assert len(jaxpr.eqns) in {
            4,
        }
        # Vars need to be primitive data types (e.g., f32) -> 2 Vars per ScaledArray
        assert jaxpr.invars[0].aval.shape == scaled_input.shape
        assert jaxpr.invars[1].aval.shape == ()
        assert jaxpr.outvars[0].aval.shape == scaled_input.shape
        assert jaxpr.outvars[1].aval.shape == ()

    def test__scalify_interpreter__with_jit__proper_jaxpr_signature(self):
        def myfunc(x):
            return x * 2

        scaled_func = scalify(jax.jit(myfunc))
        scaled_input = scaled_array([1.0, 2.0], 3, dtype=np.float32)
        jaxpr = jax.make_jaxpr(scaled_func)(scaled_input).jaxpr
        # One main jit equation.
        # TODO: additional mul in `safe_check_dtypes` mode.
        assert len(jaxpr.eqns) in {1, 2}
        eqn = jaxpr.eqns[0]
        assert eqn.primitive.name in ("pjit", "xla_call")
        assert eqn.params["name"] == "myfunc"
        # TODO: other parameters.
        # Vars need to be primitive data types (e.g., f32) -> 2 Vars per ScaledArray
        assert jaxpr.invars[0].aval.shape == scaled_input.shape
        assert jaxpr.invars[1].aval.shape == ()
        assert jaxpr.outvars[0].aval.shape == scaled_input.shape
        assert jaxpr.outvars[1].aval.shape == ()

    @chex.variants(with_jit=True, without_jit=True)
    @parameterized.parameters(
        # Identity function!
        {"fn": lambda x: x, "inputs": [scaled_array([1.0, 2.0], 3, dtype=np.float32)]},
        # Non-trivial input JAX pytree.
        {
            "fn": lambda vals: vals["x"] * vals["y"],
            "inputs": [
                {
                    "x": scaled_array([1.0, 2.0], 3, dtype=np.float32),
                    "y": scaled_array([1.5, -2.5], 2, dtype=np.float32),
                }
            ],
        },
        # Non-trivial output JAX pytree
        {"fn": lambda x: {"x": (x,)}, "inputs": [scaled_array([1.0, 2.0], 3, dtype=np.float32)]},
        # Multi-inputs operation.
        {
            "fn": lambda x, y: x * y,
            "inputs": [scaled_array([-2.0, 0.5], 0.5, dtype=np.float32), scaled_array([1.5, 1.5], 2, dtype=np.float32)],
        },
        # Proper forwarding of attributes.
        {
            "fn": lambda x: jax.lax.convert_element_type(x, np.float16),
            "inputs": [scaled_array([-4.0, 2.0], 0.5, dtype=np.float32)],
        },
        # Proper constant scalar handling.
        {
            "fn": lambda x: x * 2,
            "inputs": [scaled_array([[-2.0, 0.5]], 0.5, dtype=np.float32)],
        },
        # Internal `pjit` primitive call.
        {
            "fn": jax.jit(lambda x, y: x * y),
            "inputs": [scaled_array([-2.0, 0.5], 0.5, dtype=np.float32), scaled_array([1.5, 1.5], 2, dtype=np.float32)],
        },
        # TODO/FIXME: Proper constant Numpy array handling.
        # {
        #     "fn": lambda x: x * np.array([2.0, 3.0], dtype=np.float32),
        #     "inputs": [scaled_array([[-2.0], [0.5]], 0.5, dtype=np.float32)],
        # },
    )
    def test__scalify_decorator__proper_graph_transformation_and_result(self, fn, inputs):
        # Scalify function + (optional) jitting.
        scaled_fn = self.variant(scalify(fn))
        scaled_output = scaled_fn(*inputs)
        # Normal JAX path, without scaled arrays.
        raw_inputs = jax.tree_util.tree_map(np.asarray, inputs, is_leaf=is_scaled_leaf)
        expected_output = self.variant(fn)(*raw_inputs)

        # Do we re-construct properly the output type (i.e. handling Pytree properly)?
        if not isinstance(expected_output, (np.ndarray, Array)):
            assert type(scaled_output) is type(expected_output)

        # Check each output in the flatten tree.
        scaled_outputs_flat, _ = jax.tree_util.tree_flatten(scaled_output, is_leaf=is_scaled_leaf)
        expected_outputs_flat, _ = jax.tree_util.tree_flatten(expected_output)
        for scaled_out, exp_out in zip(scaled_outputs_flat, expected_outputs_flat):
            assert isinstance(scaled_out, ScaledArray)
            assert scaled_out.scale.shape == ()
            assert scaled_out.dtype == exp_out.dtype
            npt.assert_array_almost_equal(scaled_out, exp_out, decimal=4)

    @chex.variants(with_jit=True, without_jit=True)
    def test__scalify_decorator__promotion_broadcasted_scalar_array(self):
        def fn(sa, b):
            # Forcing broadcasting before the `lax.mul`
            b = jax.lax.broadcast_in_dim(b, sa.shape, ())
            return sa * b

        sa = scaled_array([0.5, 1.0], np.float32(4.0), dtype=np.float32)
        b = jnp.array(4.0, dtype=np.float16)

        scaled_fn = self.variant(scalify(fn))
        sout = scaled_fn(sa, b)
        expected_out = fn(np.asarray(sa), b)

        assert isinstance(sout, ScaledArray)
        # Proper output scale, with `b` treated as scaled scalar.
        npt.assert_equal(np.asarray(sout.scale), np.float32(16))
        npt.assert_array_equal(np.asarray(sout), expected_out)

    @chex.variants(with_jit=True, without_jit=True)
    def test__scalify_decorator__custom_jvp__proper_graph_transformation_and_result(self):
        # JAX official `jvp` example.
        @jax.custom_jvp
        def f(x, y):
            return jnp.sin(x) * y

        @f.defjvp
        def f_jvp(primals, tangents):
            x, y = primals
            x_dot, y_dot = tangents
            primal_out = f(x, y)
            tangent_out = jnp.cos(x) * x_dot * y + jnp.sin(x) * y_dot
            return primal_out, tangent_out

        def fn(x, y):
            return jax.jvp(f, (x, y), (x, y))

        # `scalify` on `custom_jvp` method.
        scaled_inputs = (
            scaled_array([-2.0, 0.5], 0.5, dtype=np.float32),
            scaled_array([1.5, -4.5], 2, dtype=np.float32),
        )
        scaled_primals, scaled_tangents = self.variant(scalify(fn))(*scaled_inputs)
        # JAX default/expected values
        inputs = tuple(map(asarray, scaled_inputs))
        primals, tangents = self.variant(fn)(*inputs)

        assert isinstance(scaled_primals, ScaledArray)
        assert isinstance(scaled_tangents, ScaledArray)
        npt.assert_array_almost_equal(scaled_primals, primals)
        npt.assert_array_almost_equal(scaled_tangents, tangents)

    @chex.variants(with_jit=True, without_jit=True)
    def test__scalify_decorator__custom_vjp__proper_graph_transformation_and_result(self):
        # JAX official `vjp` example.
        @jax.custom_vjp
        def f(x, y):
            return jnp.sin(x) * y

        def f_fwd(x, y):
            return f(x, y), (jnp.cos(x), jnp.sin(x), y)

        def f_bwd(res, g):
            cos_x, sin_x, y = res
            return (cos_x * g * y, sin_x * g)

        f.defvjp(f_fwd, f_bwd)

        def fn(x, y):
            primals, f_vjp = jax.vjp(f, x, y)
            return primals, f_vjp(x * y)

        # `scalify` on `custom_jvp` method.
        scaled_inputs = (
            scaled_array([-2.0, 0.5], 0.5, dtype=np.float32),
            scaled_array([1.5, -4.5], 2, dtype=np.float32),
        )
        scaled_primals, scaled_grads = self.variant(scalify(fn))(*scaled_inputs)
        # JAX default/expected values
        inputs = tuple(map(asarray, scaled_inputs))
        primals, grads = self.variant(fn)(*inputs)

        assert isinstance(scaled_primals, ScaledArray)
        npt.assert_array_almost_equal(scaled_primals, primals)
        for g, sg in zip(grads, scaled_grads):
            assert isinstance(sg, ScaledArray)
            npt.assert_array_almost_equal(sg, g)

    def test__scalify_config__default_values(self):
        cfg = get_scalify_config()
        assert isinstance(cfg, ScalifyConfig)
        assert cfg.rounding_mode == Pow2RoundMode.DOWN
        assert cfg.scale_dtype is None

    def test__scalify_config__context_manager(self):
        with ScalifyConfig(rounding_mode=Pow2RoundMode.NONE, scale_dtype=np.float32):
            cfg = get_scalify_config()
            assert isinstance(cfg, ScalifyConfig)
            assert cfg.rounding_mode == Pow2RoundMode.NONE
            assert cfg.scale_dtype == np.float32

    @chex.variants(with_jit=True, without_jit=True)
    @parameterized.parameters(
        {"scale_dtype": np.float16},
        {"scale_dtype": np.float32},
    )
    def test__scalify_config__scale_dtype_used_in_interpreter_promotion(self, scale_dtype):
        def fn(x):
            # Sub-normal "learning rate" => can create issue when converting to FP16 scaled array.
            # return x * 3.123283386230469e-05
            # FIXME: issue when using the smallest FP16 sub-normal!
            return x * (np.finfo(np.float16).smallest_subnormal * 2)

        expected_output = fn(np.float16(1))

        with ScalifyConfig(scale_dtype=scale_dtype):
            scaled_input = scaled_array(np.array(2.0, np.float16), scale=scale_dtype(0.5))
            scaled_output = self.variant(scalify(fn))(scaled_input)
            assert scaled_output.scale.dtype == scale_dtype
            npt.assert_equal(np.asarray(scaled_output, dtype=np.float32), expected_output)
